from abc import ABCMeta, abstractmethod
from typing import Iterator, TypeVar

from tdm import TalismanDocument


class AbstractReader(metaclass=ABCMeta):
    @abstractmethod
    def read(self) -> Iterator[TalismanDocument]:
        pass


_AbstractConfigurableReader = TypeVar('_AbstractConfigurableReader', bound='AbstractConfigurableReader')


class AbstractConfigurableReader(AbstractReader, metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def from_config(cls: _AbstractConfigurableReader, config) -> _AbstractConfigurableReader:
        pass



import asyncio
from abc import ABCMeta
from functools import partial
from itertools import chain
from typing import Iterator, Protocol, Sequence, Tuple, Type, TypeVar, runtime_checkable

from more_itertools import ichunked
from tdm import TalismanDocument

from tp_interfaces.abstract.model import AsyncModel
from tp_interfaces.abstract.schema import ImmutableBaseModel
from tp_interfaces.logging.context import with_log_extras
from tp_interfaces.serializable import Serializable

_Config = TypeVar('_Config', bound=ImmutableBaseModel)


@runtime_checkable
class AbstractDocumentProcessor(AsyncModel, Protocol[_Config]):

    async def process_doc(self, document: TalismanDocument, config: _Config) -> TalismanDocument:
        pass

    async def process_docs(self, documents: Sequence[TalismanDocument], config: _Config) -> Tuple[TalismanDocument, ...]:
        log_extras = with_log_extras(doc_id=lambda kwargs: kwargs['document'].id)
        return tuple(await asyncio.gather(*map(partial(log_extras(self.process_doc), config=config), documents)))

    def process_stream(self, documents: Iterator[TalismanDocument], config: _Config, batch_size: int) \
            -> Iterator[TalismanDocument]:
        return chain.from_iterable(map(partial(self.process_docs, config=config), map(tuple, ichunked(documents, batch_size))))

    @property
    def config_type(self) -> Type[_Config]:
        raise NotImplementedError


class AbstractSerializableDocumentProcessor(AbstractDocumentProcessor, Serializable, metaclass=ABCMeta):
    pass



from functools import singledispatch
from typing import Iterable

from tdm import TalismanDocument
from tdm.abstract.datamodel import AbstractDomainType, AbstractFact, AbstractNodeMention, AbstractValue, FactStatus
from tdm.datamodel.domain import AtomValueType, DocumentType
from tdm.datamodel.facts import AtomValueFact, ConceptFact, KBConceptValue, MentionFact, PropertyFact
from typing_extensions import Protocol

from tp_interfaces.domain.abstract import AbstractIdentifyingPropertyType, AbstractLiteralValueType


def get_metadata(document: TalismanDocument) -> dict[str, tuple[AbstractValue, ...] | AbstractValue]:
    document_facts = document.get_facts(ConceptFact, filter_=lambda f: isinstance(f.type_id, DocumentType))
    metadata_property_facts: set[PropertyFact] = set()
    for fact in document_facts:
        metadata_property_facts.update(document.related_facts(fact, PropertyFact))

    return {property_fact.name: property_fact.target.value for property_fact in metadata_property_facts}


class MentionedFactsFactory(Protocol):
    def __call__(
            self, mention: AbstractNodeMention,
            status: FactStatus = FactStatus.NEW,
            value: AbstractValue | Iterable[AbstractValue] = ()
    ) -> Iterable[AbstractFact]:
        ...


@singledispatch
def mentioned_fact_factory(type_: AbstractDomainType) -> MentionedFactsFactory:
    raise NotImplementedError


@mentioned_fact_factory.register(AtomValueType)
def _value_fact_factory(type_: AbstractLiteralValueType) -> MentionedFactsFactory:
    def create_facts(
            mention: AbstractNodeMention, status: FactStatus, value: tuple[AbstractValue, ...] | AbstractValue = ()
    ) -> tuple[AtomValueFact, MentionFact]:
        value_fact = AtomValueFact(status, type_, value)
        return value_fact, MentionFact(status, mention, value_fact)

    return create_facts


@mentioned_fact_factory.register(AbstractIdentifyingPropertyType)
def _concept_fact_factory(type_: AbstractIdentifyingPropertyType) -> MentionedFactsFactory:
    def create_facts(
            mention: AbstractNodeMention, status: FactStatus, value: tuple[AbstractValue, ...] | AbstractValue = ()
    ) -> tuple[ConceptFact, AtomValueFact, PropertyFact, MentionFact]:
        if status is FactStatus.APPROVED:
            concept_fact = ConceptFact(status, type_.source, KBConceptValue(value.value))
        else:
            concept_fact = ConceptFact(status, type_.source)
        value_fact = AtomValueFact(status, type_.target, value)
        return concept_fact, value_fact, PropertyFact(status, type_, concept_fact, value_fact), MentionFact(status, mention, value_fact)

    return create_facts



from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Iterable, TextIO

from tdm import TalismanDocument


class AbstractSerializer(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, doc: TalismanDocument, stream: TextIO):
        pass

    @staticmethod
    def _check_stream(stream: TextIO):
        if stream.closed or not stream.writable():
            raise Exception("stream  is closed or not writeable")


class AbstractPathSerializer(metaclass=ABCMeta):
    @abstractmethod
    def serialize(self, docs: Iterable[TalismanDocument], path: Path):
        pass



from abc import ABCMeta, abstractmethod
from contextlib import AbstractAsyncContextManager
from typing import Generic, Optional, Type, TypeVar

from tdm import TalismanDocument
from typing_extensions import Self

from tp_interfaces.abstract import ImmutableBaseModel

_LoaderConfig = TypeVar('_LoaderConfig', bound=ImmutableBaseModel)


class Loader(AbstractAsyncContextManager, Generic[_LoaderConfig], metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def from_config(cls, config: dict) -> Self:
        pass

    @property
    @abstractmethod
    def config_type(self) -> Type[_LoaderConfig]:
        pass

    @abstractmethod
    async def load_doc_and_bind_facts(self, doc: TalismanDocument, config: _LoaderConfig) -> Optional[str]:
        pass
